import os
from unittest import TestCase
from aibolit.patterns.joined_validation.joined_validation import JoinedValidation
from pathlib import Path


class TestJoinedValidation(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    pattern = JoinedValidation()

    def test_canFindSimpleJoinedValidation(self):
        file = Path(self.dir_path, 'SimpleJoinedValidation.java')
        self.assertEqual(
            [3],
            self.pattern.value(file),
            'Could not find simple joined validation'
        )

    def test_canFindJoinedValidationAndOr(self):
        file = Path(self.dir_path, 'JoinedValidationAndOr.java')
        self.assertEqual(
            [3],
            self.pattern.value(file),
            'Could not find joined validation in AndOr condition'
        )

    def test_canFindJoinedValidationOrAnd(self):
        file = Path(self.dir_path, 'JoinedValidationOrAnd.java')
        self.assertEqual(
            [3],
            self.pattern.value(file),
            'Could not find joined validation in OrAnd condition'
        )

    def test_canFindJoinedValidationOrOr(self):
        file = Path(self.dir_path, 'JoinedValidationOrOr.java')
        self.assertEqual(
            [3],
            self.pattern.value(file),
            'Could not find joined validation in OrOr condition'
        )

    def test_canFindJoinedValidationOrFunctionCall(self):
        file = Path(self.dir_path, 'JoinedValidationOrFunctionCall.java')
        self.assertEqual(
            [8],
            self.pattern.value(file),
            'Could not find joined validation in function call'
        )

    def test_canFindJoinedValidationOrFieldAccess(self):
        file = Path(self.dir_path, 'JoinedValidationOrFieldAccess.java')
        self.assertEqual(
            [6],
            self.pattern.value(file),
            'Could not find joined validation in field access'
        )

    def test_canFindNoBracketsJoinedValidation(self):
        file = Path(self.dir_path, 'NoBracketsJoinedValidation.java')
        self.assertEqual(
            [3],
            self.pattern.value(file),
            'Could not find joined validation when using no brackets'
        )

    def test_canSkipEmptyJoinedValidation(self):
        file = Path(self.dir_path, 'EmptyJoinedValidation.java')
        self.assertEqual(
            [],
            self.pattern.value(file),
            'Could not skip empty joined validation'
        )

    def test_canSkipNoJoinedValidation(self):
        file = Path(self.dir_path, 'NoJoinedValidation.java')
        self.assertEqual(
            [],
            self.pattern.value(file),
            'Could not skip when there is no joined validation'
        )



import javalang
from typing import List

from aibolit.utils.ast_builder import build_ast


class SendNull:

    def __init__(self):
        pass

    def __find_position_recursively(self, node):
        if not hasattr(node, 'children'):
            return
        else:
            for i in node.children:
                if not isinstance(i, list):
                    if hasattr(i, '_position'):
                        if i.position:
                            return i._position.line
                        else:
                            for j in i.children:
                                position = self.__find_position_recursively(j)
                                if position:
                                    return position
                else:
                    for j in i:
                        position = self.__find_position_recursively(j)

        return position

    # flake8: noqa
    # after fix addition, errors are shown
    def value(self, filename: str) -> List[int]:

        lst: List[int] = []
        tree = build_ast(filename)

        invocation_tree = tree.filter(javalang.tree.Invocation)
        arg_list = [x for _, x in invocation_tree]

        for argument in arg_list:
            ternary_list = argument.filter(javalang.tree.TernaryExpression)
            for _, expr in ternary_list:
                if isinstance(expr.if_false, javalang.tree.Literal) and expr.if_false.value == 'null':
                    if hasattr(argument, '_position'):
                        lst.append(argument._position.line)
                    else:
                        position = self.__find_position_recursively(expr)
                        lst.append(position)
                if isinstance(expr.if_true, javalang.tree.Literal) and expr.if_true.value == 'null':
                    if hasattr(argument, '_position'):
                        lst.append(argument._position.line)
                    else:
                        position = self.__find_position_recursively(expr)
                        lst.append(position)

        for _, node in tree:
            try:
                for argument in node.arguments:
                    if isinstance(argument, javalang.tree.Literal) and argument.value == "null" and \
                            argument._position.line not in lst:
                        lst.append(argument._position.line)
            except (AttributeError, TypeError):
                pass
        lst = sorted(lst)
        return lst



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List, Iterator

from networkx import DiGraph  # type: ignore

from aibolit.ast_framework._auxiliary_data import common_attributes, attributes_by_node_type, ASTNodeReference


class ASTNode:
    def __init__(self, graph: DiGraph, node_index: int):
        self._graph = graph
        self._node_index = node_index

    @property
    def children(self) -> Iterator['ASTNode']:
        for child_index in self._graph.succ[self._node_index]:
            yield ASTNode(self._graph, child_index)

    @property
    def node_index(self) -> int:
        return self._node_index

    def __getattr__(self, attribute_name: str):
        if attribute_name not in common_attributes:
            node_type = self._graph.nodes[self._node_index]['node_type']
            if(attribute_name not in attributes_by_node_type[node_type]):
                raise AttributeError(f'{node_type} node does not have "{attribute_name}" attribute.')

        attribute = self._graph.nodes[self._node_index][attribute_name]
        if isinstance(attribute, ASTNodeReference):
            attribute = ASTNode(self._graph, attribute.node_index)
        elif isinstance(attribute, list) and \
                all((isinstance(item, ASTNodeReference) for item in attribute)):
            attribute = [ASTNode(self._graph, item.node_index) for item in attribute]
        return attribute

    def __dir__(self) -> List[str]:
        node_type = self._graph.nodes[self._node_index]['node_type']
        return ASTNode._public_fixed_interface + \
            list(common_attributes) + list(attributes_by_node_type[node_type])

    def __str__(self) -> str:
        text_representation = f'node index: {self._node_index}'
        node_type = self.__getattr__('node_type')
        for attribute_name in sorted(common_attributes | attributes_by_node_type[node_type]):
            text_representation += f'\n{attribute_name}: {self.__getattr__(attribute_name)}'

        return text_representation

    def __repr__(self) -> str:
        return f'<ASTNode node_type: {self.__getattr__("node_type")}, node_index: {self._node_index}>'

    # names of methods and properties, which is not generated dynamically
    _public_fixed_interface = ['children', 'node_index', 'subtree']



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List

import javalang.ast
import javalang.parse
import javalang.tree

from aibolit.types_decl import LineNumber
from aibolit.utils.ast_builder import build_ast


class ManyPrimaryCtors(object):
    def value(self, filename: str):
        tree = build_ast(filename)

        return self.__traverse_node(tree)

    def __traverse_node(self, tree: javalang.ast.Node) -> List[LineNumber]:
        lines: List[LineNumber] = list()

        for _, class_declaration in tree.filter(javalang.tree.ClassDeclaration):
            primary_ctors = list(filter(_is_primary, class_declaration.constructors))

            if len(primary_ctors) > 1:
                lines.extend(ctor.position.line for ctor in primary_ctors)

        return lines


def _is_primary(constructor: javalang.tree.ConstructorDeclaration) -> bool:
    for _, assignment in constructor.filter(javalang.tree.Assignment):
        if _is_instance_variable_assignment(assignment):
            return True

    return False


def _is_instance_variable_assignment(assignment: javalang.tree.Assignment) -> bool:
    return isinstance(assignment.expressionl, javalang.tree.This)



import javalang
from typing import List
import uuid
from collections import defaultdict
import hashlib
import itertools
from javalang.tree import FormalParameter

from aibolit.utils.ast_builder import build_ast


class MultipleTry:

    def __init__(self):
        pass

    def traverse_node(self, node, dict_with_chains, uuid_method):
        if not node:
            return dict_with_chains

        for item in node.children:
            if item and (isinstance(item, tuple) or isinstance(item, list)):
                for j in item:
                    if isinstance(j, javalang.tree.MethodInvocation):
                        if not j.qualifier and j.qualifier != '':
                            # it means that there is method chaining
                            dict_with_chains[uuid_method].append([j.position.line, j.member])
                            self.traverse_node(j, dict_with_chains, uuid_method)
                        else:
                            # it means that we have method invocation without chaining like
                            # result.add(field.getName(), column.columnName(), field.getType());
                            new_uuid = uuid.uuid1()
                            dict_with_chains[new_uuid].append([j.position.line, j.member])
                            self.traverse_node(j, dict_with_chains, new_uuid)

                    elif isinstance(j, javalang.tree.MethodDeclaration):
                        self.traverse_node(j, dict_with_chains, str(uuid.uuid1()))

                    elif isinstance(j, javalang.tree.StatementExpression):
                        self.traverse_node(j, dict_with_chains, uuid_method)

                    elif isinstance(j, javalang.tree.This) or isinstance(j, javalang.tree.ClassCreator):
                        self.traverse_node(j, dict_with_chains, str(uuid.uuid1()))
            elif isinstance(item, javalang.tree.ClassCreator):
                self.traverse_node(item, dict_with_chains, uuid_method)

        return dict_with_chains


    # flake8: noqa: C901
    def value(self, filename: str):
        """
        Travers over AST tree and fins function with nested/sequential try statement
        :param filename:
        :return:
        List of tuples with LineNumber and List of methods names, e.g.
        [[10, 'func1'], [10, 'fun2']], [[23, 'run'], [23, 'start']]]
        """

        tree = build_ast(filename)
        res = defaultdict(list)
        for _, method_node in tree.filter(javalang.tree.MethodDeclaration):
            for _, try_node in method_node.filter(javalang.tree.TryStatement):
                formal_params = [
                    (x.type.name + ' ' + x.name)  
                    for x in method_node.parameters 
                    if isinstance(x, FormalParameter)
                ]
                func_name = '{f}({params})'.format(
                    f=method_node.name,
                    params=','.join(formal_params)
                ).encode('utf-8')
                m = hashlib.md5()
                m.update(func_name)
                res[m.hexdigest()].append(method_node.position.line)

        return list(set(itertools.chain.from_iterable([y for x, y in res.items() if len(y) > 1])))


from aibolit.utils.ast_builder import build_ast
from aibolit.ast_framework import AST, ASTNodeType
from itertools import islice
from aibolit.ast_framework.java_package import JavaPackage


class FanOut:
    '''
    Fan Out metric is defined as the number of other classes referenced by a class.
    '''
    def __init__(self):
        pass

    def value(self, filename: str) -> int:  # noqa: C901

        # exception are used from https://checkstyle.sourceforge.io/config_metrics.html#ClassFanOutComplexity
        considered_classes = {'ArrayIndexOutOfBoundsException': 0, 'ArrayList': 0, 'Boolean': 0, 'Byte': 0,
                              'Character': 0, 'Class': 0, 'Deprecated': 0, 'Deque': 0, 'Double': 0,
                              'Exception': 0, 'Float': 0, 'FunctionalInterface': 0, 'HashMap': 0,
                              'HashSet': 0, 'IllegalArgumentException': 0, 'IllegalStateException': 0,
                              'IndexOutOfBoundsException': 0, 'Integer': 0, 'LinkedList': 0, 'List': 0,
                              'Long': 0, 'Map': 0, 'NullPointerException': 0, 'Object': 0, 'Override': 0,
                              'Queue': 0, 'RuntimeException': 0, 'SafeVarargs': 0, 'SecurityException': 0,
                              'Set': 0, 'Short': 0, 'SortedMap': 0, 'SortedSet': 0, 'String': 0, 'StringBuffer': 0,
                              'StringBuilder': 0, 'SuppressWarnings': 0, 'Throwable': 0, 'short': 0, 'void': 0,
                              'TreeMap': 0, 'TreeSet': 0, 'UnsupportedOperationException': 0, 'Void': 0,
                              'System.out': 0, 'boolean': 0, 'byte': 0, 'char': 0, 'double': 0, 'float': 0,
                              'int': 0, 'long': 0,
                              }
        fan_outs = 0

        # check imported classes
        tree = AST.build_from_javalang(build_ast(filename))
        for each_import in (tree.children_with_type(tree.root, ASTNodeType.IMPORT)):
            name_node, = islice(tree.children_with_type(each_import, ASTNodeType.STRING), 1)
            new_class = tree.get_attr(name_node, 'string').split('.')[-1]
            if considered_classes.get(new_class) is None:
                fan_outs += 1
                considered_classes[new_class] = 0

        p = JavaPackage(filename)
        for class_name in p.java_classes:
            tree = p.java_classes[class_name]
            for var_node in tree.get_nodes(ASTNodeType.VARIABLE_DECLARATOR):
                var_child = list(tree.children_with_type(var_node, ASTNodeType.STRING))
                new_class_name = tree.get_attr(var_child[0], 'string')

                for class_creator_node in tree.children_with_type(var_node, ASTNodeType.CLASS_CREATOR):
                    for go_to_name in tree.children_with_type(class_creator_node, ASTNodeType.REFERENCE_TYPE):
                        classC_child = list(tree.children_with_type(go_to_name, ASTNodeType.STRING))
                        used_class_name = tree.get_attr(classC_child[0], 'string')
                        if considered_classes.get(used_class_name) is None:
                            considered_classes[used_class_name] = 0
                            fan_outs += 1
                        if considered_classes.get(new_class_name) is None:
                            considered_classes[new_class_name] = 0

        # check classes of invokated methods
        for i in tree.get_nodes(ASTNodeType.STATEMENT_EXPRESSION):
            for invoked_method_child in tree.children_with_type(i, ASTNodeType.METHOD_INVOCATION):
                name_of_invoked_class = tree.get_method_invocation_params(invoked_method_child)
                if considered_classes.get(name_of_invoked_class.object_name) is None:
                    considered_classes[name_of_invoked_class.object_name] = 0
                    fan_outs += 1

        return fan_outs



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List

from aibolit.utils.ast_builder import build_ast
from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class JoinedValidation:
    '''
    Finds all if statements, which "then" branch consist of a single throw statement
    and logical or "||" used in condition.
    '''

    def value(self, filename: str) -> List[int]:
        ast = AST.build_from_javalang(build_ast(filename))
        lines: List[int] = []
        for if_statement in ast.get_proxy_nodes(ASTNodeType.IF_STATEMENT):
            if self._is_logical_or_used_in_expression(if_statement.condition) and \
               self._is_block_consist_of_single_throw(if_statement.then_statement):
                lines.append(if_statement.line)
        return lines

    def _is_logical_or_used_in_expression(self, expression: ASTNode) -> bool:
        if expression.node_type == ASTNodeType.BINARY_OPERATION and expression.operator == '||':
            return True

        return any(self._is_logical_or_used_in_expression(child) for child in expression.children)

    def _is_block_consist_of_single_throw(self, block: ASTNode) -> bool:
        if block.node_type == ASTNodeType.THROW_STATEMENT:
            return True

        children = list(block.children)
        return len(children) == 1 and children[0].node_type == ASTNodeType.THROW_STATEMENT



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from networkx import dfs_labeled_edges  # type: ignore

from aibolit.utils.ast_builder import build_ast
from aibolit.ast_framework import AST, ASTNodeType


class NCSSMetric():
    def __init__(self):
        pass

    def value(self, filename: str):
        set_of_the_types = {ASTNodeType.ANNOTATION_DECLARATION,
                            ASTNodeType.CLASS_DECLARATION,
                            ASTNodeType.CONSTANT_DECLARATION,
                            ASTNodeType.CONSTRUCTOR_DECLARATION,
                            ASTNodeType.DECLARATION,
                            ASTNodeType.ENUM_CONSTANT_DECLARATION,
                            ASTNodeType.ENUM_DECLARATION,
                            ASTNodeType.FIELD_DECLARATION,
                            ASTNodeType.INTERFACE_DECLARATION,
                            ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                            ASTNodeType.METHOD_DECLARATION,
                            ASTNodeType.TYPE_DECLARATION,
                            ASTNodeType.VARIABLE_DECLARATION,
                            ASTNodeType.CATCH_CLAUSE,
                            ASTNodeType.ASSERT_STATEMENT,
                            ASTNodeType.BREAK_STATEMENT,
                            ASTNodeType.CONTINUE_STATEMENT,
                            ASTNodeType.DO_STATEMENT,
                            ASTNodeType.FOR_STATEMENT,
                            ASTNodeType.IF_STATEMENT,
                            ASTNodeType.RETURN_STATEMENT,
                            ASTNodeType.STATEMENT,
                            ASTNodeType.STATEMENT_EXPRESSION,
                            ASTNodeType.SWITCH_STATEMENT,
                            ASTNodeType.SWITCH_STATEMENT_CASE,
                            ASTNodeType.SYNCHRONIZED_STATEMENT,
                            ASTNodeType.THROW_STATEMENT,
                            ASTNodeType.TRY_STATEMENT,
                            ASTNodeType.WHILE_STATEMENT}

        if len(filename) == 0:
            raise ValueError('Empty file for analysis')

        tree = AST.build_from_javalang(build_ast(filename))
        metric = 0
        for _, destination, edge_type in dfs_labeled_edges(tree.tree, tree.root):
            if edge_type == 'forward':
                node_type = tree.get_type(destination)
                if node_type in set_of_the_types:
                    metric += 1
                if node_type == ASTNodeType.FOR_CONTROL:
                    metric -= len(list(tree.children_with_type(destination, ASTNodeType.VARIABLE_DECLARATION)))
                    metric -= len(list(tree.children_with_type(destination, ASTNodeType.LOCAL_VARIABLE_DECLARATION)))

        return metric



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import List, Optional, Tuple, Dict

import javalang

from aibolit.utils.java_parser import JavalangImproved, ASTNode


class VarDeclarationDistance:
    '''
    Returns lines where variable first time used but declared more than
    specific number of lined before
    '''

    def __init__(self, lines_th: int):
        self.__lines_th = lines_th

    def __node_name(self, node) -> Optional[str]:
        qualifier = node.qualifier if hasattr(node, 'qualifier') else None
        member = node.member if hasattr(node, 'member') else None
        name = node.name if hasattr(node, 'name') else None
        return qualifier or member or name

    def __group_vars_by_method(self, items: List[Tuple[ASTNode, Optional[str]]]) -> List[Dict]:
        '''
        Group variables by method scope and calculate for each the declaration
        line and first usage line
        '''
        var_scopes: List[Dict] = []
        vars: Dict = {}
        unique_methods = list(set(
            map(lambda v: v[0].method_line, items)
        ))

        for method in unique_methods:
            method_items = map(
                lambda v: {"line": v[0].line, "name": v[1], "ntype": type(v[0].node)},
                filter(lambda v: v[0].method_line == method, items)
            )
            vars = {}
            var_scopes += [vars]
            for item in method_items:
                if item['ntype'] in [javalang.tree.MethodDeclaration]:
                    vars = {}
                    var_scopes += [vars]
                    continue

                if (item['ntype'] == javalang.tree.VariableDeclarator):
                    vars[item['name']] = {'decl': item['line'], 'first_usage': None}
                    continue

                if (item['ntype'] == javalang.tree.VariableDeclarator):
                    vars[item['name']] = {'decl': item['line']}
                    continue

                if item['name'] in vars.keys() and vars[item['name']]['first_usage'] is None:
                    vars[item['name']]['first_usage'] = item['line']

        return var_scopes

    def __line_diff(self, usage_line: int, declaration_line: int, empty_lines: List[int]) -> int:
        '''
        Calculate line difference between variable declaration and first usage
        taking into account empty lines
        '''
        lines_range = set(range(declaration_line + 1, usage_line))
        return len(lines_range.difference(empty_lines))

    def value(self, filename: str) -> List[int]:
        ''''''
        tree = JavalangImproved(filename)
        empty_lines = tree.get_empty_lines()
        items = list(
            map(lambda v: (v, self.__node_name(v.node)), tree.tree_to_nodes())
        )
        var_scopes = self.__group_vars_by_method(items)
        violations = []
        for scope in var_scopes:
            for var in scope:
                if scope[var]['first_usage'] is None:
                    continue

                line_diff = self.__line_diff(
                    scope[var]['first_usage'],
                    scope[var]['decl'],
                    empty_lines
                )
                if line_diff < self.__lines_th:
                    continue

                violations.append(scope[var]['first_usage'])

        return violations




import os
from unittest import TestCase
from aibolit.patterns.private_static_method.private_static_method import PrivateStaticMethod
from pathlib import Path


class TestPrivateStaticMethod(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_find_private_static_methods(self):
        self.assertEqual([2], PrivateStaticMethod().value(Path(self.dir_path, 'PrivateStaticMethod.java')))



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.var_middle.var_middle import VarMiddle


class VarMiddleTest(TestCase):
    def test_good_class(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/1.java')
        self.assertEqual(lines, [])

    def test_bad_class(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/2.java')
        self.assertEqual(lines, [9, 16])

    def test_case_with_multiline_method_declaration(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/3.java')
        self.assertEqual(lines, [])

    def test_case_with_empty_lines(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/4.java')
        self.assertEqual(lines, [])

    def test_case_autoclosable(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/5.java')
        self.assertEqual(lines, [])

    def test_case_nested_class(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/6.java')
        self.assertEqual(lines, [30, 33])

    def test_declaration_after_super_class_method_call(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/7.java')
        self.assertEqual(lines, [14])

    def test_for_scope_good(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/8.java')
        self.assertEqual(lines, [])

    def test_for_scope_bad(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/9.java')
        self.assertEqual(lines, [11])

    def test_variable_declared_after_for(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/10.java')
        self.assertEqual(lines, [11])

    def test_11(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/11.java')
        self.assertEqual(lines, [])

    def test_catch_good(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/12.java')
        self.assertEqual(lines, [])

    def test_catch_bad(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/13.java')
        self.assertEqual(lines, [38])

    def test_else_bad(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/14.java')
        self.assertEqual(lines, [88])

    def test_variable_after_curly_braces(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/15.java')
        self.assertEqual(lines, [])

    def test_variable_inside_lambda(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/16.java')
        self.assertEqual(lines, [])

    def test_annotation_with_parameters(self):
        pattern = VarMiddle()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/17.java')
        self.assertEqual(lines, [22])



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from statistics import variance

from aibolit.utils.encoding_detector import read_text_with_autodetected_encoding
from aibolit.utils.utils import RemoveComments


class IndentationCounter:

    def __init__(self, left_var=False, right_var=False, max_left=False, max_right=False):
        self.left_var = left_var
        self.right_var = right_var
        self.max_left = max_left
        self.max_right = max_right

    def __file_to_tokens(self, filename: str):
        """
        Takes path to java class file and returns tokens

        :param filename: file name
        :return: list of counted spaces
        """
        source_code = read_text_with_autodetected_encoding(filename)
        source_code = RemoveComments.remove_comments(source_code)
        return [line.replace('\t', '    ') for line in source_code.splitlines()
                if line]

    def value(self, filename: str):
        lines = self.__file_to_tokens(filename)
        if not lines or len(lines) == 1:
            return 0

        spaces_per_line = []
        prev_left = 0
        prev_right = len(lines[0])
        for i, line in enumerate(lines):
            first_non_space_symbol_pos = len(line) - len(line.lstrip(' '))
            spaces_per_line.append(
                [i,
                 # count the difference between the position of first symbol in cur string and
                 # the position of first symbol of previous string
                 first_non_space_symbol_pos - prev_left,
                 # count the difference between the position of last symbol in cur string and
                 # the position of last symbol of previous string
                 len(line) - prev_right
                 ])
            prev_left = first_non_space_symbol_pos
            prev_right = len(line)

        val = None
        if self.left_var:
            val = variance([x[1] for x in spaces_per_line])
        elif self.right_var:
            val = variance([x[2] for x in spaces_per_line])
        elif self.max_left:
            val = max([abs(x[1]) for x in spaces_per_line])
        elif self.max_right:
            val = max([abs(x[2]) for x in spaces_per_line])

        return val



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List, Dict, Set, Iterator

from networkx import DiGraph, strongly_connected_components, weakly_connected_components  # type: ignore

from aibolit.ast_framework import AST
from aibolit.ast_framework.java_class import JavaClass


def decompose_java_class(java_class: JavaClass, strength: str) -> List[AST]:
    '''
    Splits java_class fields and methods by their usage and
    construct for each case an AST with only those fields and methods kept.
    Use "strength" parameter to control splitting criteria. Use "strong" or "weak"
    for splitting fields and methods by strong and weak connectivity.
    '''
    usage_graph = DiGraph()
    fields_ids: Dict[str, int] = {}
    methods_ids: Dict[str, int] = {}
    for field_name in java_class.fields:
        fields_ids[field_name] = len(fields_ids)
        usage_graph.add_node(fields_ids[field_name], type='field', name=field_name)
    for method_name in java_class.methods:
        methods_ids[method_name] = len(fields_ids) + 1 + len(methods_ids)
        usage_graph.add_node(methods_ids[method_name], type='method', name=method_name)

    for method_name in java_class.methods:
        for method in java_class.methods[method_name]:
            for used_field_name in method.used_fields:
                usage_graph.add_edge(methods_ids[method_name], fields_ids[used_field_name])
            for used_method_name in method.used_methods:
                usage_graph.add_edge(methods_ids[method_name], methods_ids[used_method_name])

    components: Iterator[Set[int]]
    if strength == 'strong':
        components = strongly_connected_components(usage_graph)
    elif strength == 'weak':
        components = weakly_connected_components(usage_graph)
    else:
        raise ValueError(f'"strength" argument must be either "strong" or "weak", but "{strength}" was provided.')

    class_parts: List[AST] = []
    for component in components:
        field_names = [usage_graph.nodes[node]['name'] for node in component
                       if usage_graph.nodes[node]['type'] == 'field']
        method_names = [usage_graph.nodes[node]['name'] for node in component
                        if usage_graph.nodes[node]['type'] == 'method']
        class_parts.append(_filter_class_methods_and_fields(java_class, field_names, method_names))

    return class_parts


def _filter_class_methods_and_fields(java_class: JavaClass, allowed_fields_names: List[str],
                                     allowed_methods_names: List[str]) -> AST:
    allowed_nodes = {java_class.root}
    for field_name in allowed_fields_names:
        field = java_class.fields[field_name]
        allowed_nodes |= field.tree.nodes
    for method_name in allowed_methods_names:
        for method in java_class.methods[method_name]:
            allowed_nodes |= method.tree.nodes

    return AST(java_class.tree.subgraph(allowed_nodes), java_class.root)




class BidirectIndex:

    def __init__(self):
        pass

    def value(self, filename: str):
        """
        Finds if a variable is being incremented and decremented within the same method
        :param filename: filename to be analyzed
        :return: list of LineNumber with the variable declaration lines
        @todo #139:30min Implement bidirect index pattern
         If the same numeric variable is incremented and decremented within the same method,
         it's a pattern. A numeric variable should either be always growing or decreasing.
         Bi-directional index is confusing. The method must return a list with the line numbers
         of the variables that match this pattern. After implementation, activate tests in
         test_bidirect_index.py
        """
        return []



import javalang

from aibolit.utils.ast_builder import build_ast


class PublicStaticMethod:

    def __init__(self):
        pass

    def value(self, filename: str):
        tree = build_ast(filename).filter(javalang.tree.MethodDeclaration)
        return [
            node.position.line for
            path, node in tree if
            all(elem in node.modifiers for elem in ['public', 'static'])
        ]



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from cached_property import cached_property  # type: ignore
from deprecated import deprecated  # type: ignore

from typing import Dict, Set, TYPE_CHECKING
from networkx import DiGraph  # type: ignore

from aibolit.ast_framework import AST, ASTNodeType
from aibolit.ast_framework.java_class_method import JavaClassMethod
from aibolit.ast_framework.java_class_field import JavaClassField

if TYPE_CHECKING:
    from aibolit.ast_framework.java_package import JavaPackage


@deprecated("This functionality must be transmitted to ASTNode")
class JavaClass(AST):
    def __init__(self, tree: DiGraph, root: int, java_package: 'JavaPackage'):
        self.tree = tree
        self.root = root
        self._java_package = java_package

    @cached_property
    def name(self) -> str:
        try:
            class_name = next(self.children_with_type(self.root, ASTNodeType.STRING))
            return self.tree.nodes[class_name]['string']
        except StopIteration:
            raise ValueError("Provided AST does not has 'STRING' node type right under the root")

    @property
    def package(self) -> 'JavaPackage':
        return self._java_package

    @cached_property
    def methods(self) -> Dict[str, Set[JavaClassMethod]]:
        methods: Dict[str, Set[JavaClassMethod]] = {}
        for method_ast in self.get_subtrees(ASTNodeType.METHOD_DECLARATION):
            method = JavaClassMethod(method_ast.tree, method_ast.root, self)
            if method.name in methods:
                methods[method.name].add(method)
            else:
                methods[method.name] = {method}
        return methods

    @cached_property
    def fields(self) -> Dict[str, JavaClassField]:
        fields: Dict[str, JavaClassField] = {}
        for field_ast in self.get_subtrees(ASTNodeType.FIELD_DECLARATION):
            field = JavaClassField(field_ast.tree, field_ast.root, self)
            fields[field.name] = field
        return fields
